# News20 example (binary classification)

In this example we will train a Logistic Regression model on the news20 dataset, using both sparkML and the spark distributed version of snapML `snap-ml-spark`. 

The news20 dataset is avaliable in the [LIBSVM repository](https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html). 

The problem is described as follows:

> This is a classification problem to classify news articles to positive and negative.

This dataset is a binary text classification dataset. It is a size-balanced two-class variant
of the UCI - Twenty Newsgroups dataset (Blake and Merz, 1998). The positive
class consists of the 10 groups with names of formsci.*, comp.*, or misc.forsale,
and the negative class consists of the other 10 groups. The dataset consists of 20K
training examples and 1.3M features.

## Download the Data

Download and decompress the data from the LIBSVM repository:
```bash
mkdir data
cd data
wget https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/news20.binary.bz2
bunzip2 news20.binary.bz2
cd ../
```

## Data Pre-processing
The dataset is highly sparse and thus it is provided in LIBSVM format. However we need to perform some additional preprocessing on the data before training our models.

First, we load the data from SvmLight format:
```python
from sklearn.datasets import load_svmlight_file
filename = "data/news20.binary"
X,y = load_svmlight_file(filename)
```

At this point we replace all the y = -1 labels to y = 0 since in the following steps we will train our model using both the snap-ml-spark and spark ML lib libraries. snap-ml-spark is compatible with all the binary classification representations (it supports both y = -1 and y = 0 as a negative class) while spark ML lib allows only y = 0. Thus we perform the following replacement:
```python
y[y == -1] = 0
```

Next, we make a train/test split of the data:
```python
from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, random_state=42)
```

Then, we normalize (by column) both the train and test data:
```python
from sklearn.preprocessing import normalize
X_train = normalize(X_train, axis=0, norm='l1')
X_test  = normalize(X_test,  axis=0, norm='l1')
```

Finally, we write the divided Train/Test datasets back to LIBSVM format.
The produced datasets are the following: (data/news20.binary-train.py, data/news20.binary-test.py)
Note that the indices start from 1 instead of 0 (`zero_based=False`).
```python
from sklearn.datasets import dump_svmlight_file
dump_svmlight_file(X_train,y_train, filename+'-train.libsvm', zero_based=False)
dump_svmlight_file(X_test,y_test, filename+'-test.libsvm', zero_based=False)
```

The entire pre-processing code can be run by executing the following python script
```bash
python preprocess-news20.py
```

## Training and Evaluating a Logistic Regression Model

Before starting with training and evaluating our Logistic Regression model, we firstly need to configure the
pyspark configuration based on our cluster setup. In this example, we run our examples on a local master
in standalone mode using 4 processes. To configure pyspark we perform the following operations in python:
```python
from pyspark import SparkConf, SparkContext
from pyspark.sql.session import SparkSession

conf = SparkConf()\
.set("spark.default.parallelism",args.num_partitions)\
.set("spark.cores.max",args.num_partitions)\
.setAppName("Spark-criteoKaggle2014-SparkComparison")\
.set("spark.executor.extraJavaOptions","-XX:MaxDirectMemorySize=8192m")\
.set("spark.driver.maxResultSize",8192)
# Create spark config
sc = SparkContext(conf=conf)
# Configure Spark
spark = SparkSession(sc)
```

Please note that the number of executors cannot overcome the number of available GPUs
when the `use_gpu=True` as we will see later in the LogisticRegression constructor.

In addition, we disable the spark logging INFO so that we can read the output of our example in an easier way (optional step).
```python
log4j = sc._jvm.org.apache.log4j
log4j.LogManager.getRootLogger().setLevel(log4j.Level.ERROR)
```

Now, after running the data pre-processing steps above, we can load the data into memory:
```python
from snap_ml_spark import DatasetReader
from snap_ml_spark import LogisticRegression as snapml_LogisticRegression
from snap_ml_spark.Metrics import accuracy
train_filename = 'data/news20.binary-train.libsvm'
test_filename = 'data/news20.binary-test.libsvm'
n_features_ = 1355191
trainData = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(train_filename)
testData = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(test_filename)
```
Note that we set the maximum index so that we make sure that we preserve the right dimensions on the dataset
since we have performed a split operation in order to create our Train and Test datasets.

Now we can create a Logistic Regression model from snap-ml-spark. We set the snapml LR regularizer to 100.0, the number of training epochs to 10 and we optimize the dual objective function with the GPUs enabled:
```python
snapml_regularizer = 100.0
snapml_lr = snapml_LogisticRegression(max_iter=10, regularizer=snapml_regularizer, verbose=False, dual=True, use_gpu=True)
```
We disable the training history since we are planning to time the execution of the training phase, we enable the GPUs and we
set the number of iterations to 10 and the regularization parameter to 100.0.

Train the model and time how long it takes:
```python
import time
snapml_t0 = time.time()
snapml_lr.fit(trainData)
snapml_time = time.time() - snapml_t0
```
Since we have disabled the training history (`verbose=False`), one might want to just check the final training loss value.
`snaml_lr.fit(trainData)` always returns this value.

We use the model to generate predictions on the test set (inference) and we store the predicted labels:
```python
pred = snapml_lr.predict(testData)
```
As a last step, we evaluate the accuracy on the test set:
```python
snapml_accuracy = accuracy(pred)
```
Now in order to fit the sparkML algorithm on the same data, we have to read them using the spark loaders:
```python
# Load training data
trainData = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(train_filename)
testData = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(test_filename)
```
Now we can create a Logistic Regression model from sparkML library. We disable the feature scaling and standardization option and we set the regularizer to 1/N where N is the number of training examples (based on the sparkML documentation snapml regularizer N is equivalent to 1/N in sparkML):
```python
from pyspark.ml.classification import LogisticRegression as sparkmllib_LogisticRegression
n_examples = trainData.count()
sparkmllib_lr = sparkmllib_LogisticRegression(fitIntercept=False, standardization=False, regParam=snapml_regularizer/n_examples)
```

We train the model and time how long it takes
```python
sparkmllib_t0 = time.time()
sparkmllib_lr_model = sparkmllib_lr.fit(trainData)
sparkmllib_time = time.time() - sparkmllib_t0
```
We perform the inference step:
```python
predictions = sparkmllib_lr_model.transform(testData)
```
And finally, we can evaluate the accuracy on the test set:
```python
from pyspark.ml.evaluation import MulticlassClassificationEvaluator
evaluator = MulticlassClassificationEvaluator(
    labelCol="label", predictionCol="prediction", metricName="accuracy")
sparkmllib_accuracy = evaluator.evaluate(predictions)
```

Then we print off the results:
```python
print('snapML accuracy: %.4f' %snapml_accuracy, ", time: %.2f" % snapml_time, 's')
print('sparkML lib accuracy: %.4f' % sparkmllib_accuracy, ", time: %.2f" % sparkmllib_time, 's')
```

One can run the entire example by executing the following commands:
```
spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-news20/example-news20.py --data_path path/to/data

Note: By default the location of --data_path is current directory. Parent directory of the directory 'data/' is to be given as parameter to --data_path option.

Command to run example using GPUs:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-news20/example-news20.py --use_gpu

Command to run example with specified number of partitions:

spark-submit --master spark://node1.ibm.com:7077 --jars $CONDA_PREFIX/snap-ml-spark/lib/snap-ml-spark-v*-ppc64le.jar $CONDA_PREFIX/snap-ml-spark/examples/example-news20/example-news20.py --num_partitions

Note: Default value of num_partitions is 4. num_partitions can be 8, if user wants to use 8 GPUs.
```
Please read the outer README.md to see how to download the snap-ml-spark.jar if you have not already done it.

One should see the the following output on a power9 architecture with 4 V100 GPUs:
```
[...]
snapML accuracy: 0.9660 , time: 12.50 s
sparkML lib accuracy: 0.9640 , time: 78.17 s
```

Please not that times can vary depending on system architecture.

&copy; Copyright IBM Corporation 2018, 2019
