# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from __future__ import print_function
import time
import argparse
import os
## SPARK

from pyspark import SparkConf, SparkContext
from pyspark.sql.session import SparkSession
defaultPath = os.getcwd()
CLI=argparse.ArgumentParser()
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--skip_sparkml",
    action = 'store_true',
    default=False
)
CLI.add_argument(
   "--data_path",
   type=str,
   default=defaultPath
)
CLI.add_argument(
   "--num_partitions",
   type=int,
   default= 4
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
args = CLI.parse_args()

if args.use_balanced_class_weights:
  class_weights = "balanced"
else:
  class_weights = None
# Set your own master (e.g 'spark://zhcc011.zurich.ibm.com:7077') or use 'local'
# Configure Spark (4 executors, local standalone mode)
# If one wants to run spark on a cluster, use your own spark configuration (set master , appName, etc...) 
conf = SparkConf()\
.set("spark.default.parallelism",args.num_partitions)\
.set("spark.cores.max",args.num_partitions)\
.setAppName("Spark-Webspam-SparkComparison")\
.set("spark.executor.extraJavaOptions","-XX:MaxDirectMemorySize=8192m")\
.set("spark.driver.maxResultSize",8192)
# Create spark config
sc = SparkContext(conf=conf)
# Configure Spark
spark = SparkSession(sc)

# disable spark INFO logging - comment out to enable
log4j = sc._jvm.org.apache.log4j
log4j.LogManager.getRootLogger().setLevel(log4j.Level.ERROR)

# data paths
filename = args.data_path + '/data/webspam_wc_normalized_trigram.svm'
train_filename_sparse_snap = filename + '-train_sparse.snap'
train_filename_sparse_snap_t = filename + '-train_sparse.snap.t'
train_filename = filename + '-train.libsvm'
test_filename = filename + '-test.libsvm'

## snapML

from snap_ml_spark import DatasetReader

from snap_ml_spark import LogisticRegression as snapml_LogisticRegression
from snap_ml_spark.Metrics import accuracy, logisticLoss

n_features_ = 16609143

train_data_t = DatasetReader().setFormat("snap").load(train_filename_sparse_snap_t)

# Load test data
print('Reading test data (snap ML)')
test_data = DatasetReader().setFormat("libsvm").setNumFt(n_features_).load(test_filename)

# Create snapML Logistic Regression - primal CPU case
snapml_regularizer = 1.0
if args.use_gpu:
    snapml_lr = snapml_LogisticRegression(max_iter=70, regularizer=snapml_regularizer, verbose=False, dual=False, use_gpu=args.use_gpu, class_weights=class_weights) 
else:
    snapml_lr = snapml_LogisticRegression(max_iter=19, regularizer=snapml_regularizer, verbose=False, dual=False, use_gpu=args.use_gpu, class_weights=class_weights)
# Fit the model and time it
snapml_t0 = time.time()
snapml_lr.fit(train_data_t)
snapml_time = time.time() - snapml_t0

# Perform inference on test data
pred = snapml_lr.predict(test_data)

# Compute accuracy
snapml_accuracy = accuracy(pred)

# Print off SnapML result
print('snapML  accuracy: %.4f' %snapml_accuracy, ", time: %.2f" % snapml_time, 's')




## SPARK ml

# Skip if --skip_sparkml flag is given
if args.skip_sparkml:
    print('Execution using Spark ML is skipped.')
    exit()

from pyspark.ml.classification import LogisticRegression as sparkml_LogisticRegression

train_filename = "file://" + train_filename
test_filename = "file://" + test_filename

# Load training data
print('Reading train data (spark ML)')
train_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(train_filename)
print('Reading test data (spark ML)')
test_data = spark.read.format("libsvm").option("numFeatures", str(n_features_)).load(test_filename)

print('counting examples...')
n_examples = train_data.count()
print("N Examples:",n_examples)

# Create sparkML lib Logistic Regression
sparkml_lr = sparkml_LogisticRegression(fitIntercept=False, regParam=snapml_regularizer/n_examples, standardization=False)

# Fit the model and time it
print('fitting spark ML model')
sparkml_t0 = time.time()
sparkml_lr_model = sparkml_lr.fit(train_data)
sparkml_time = time.time() - sparkml_t0

# Perform inference on test data
predictions = sparkml_lr_model.transform(test_data)

# Show predictions against test labels
# predictions.select("rawPrediction", "prediction", "label", "features").show(10)

# Compute accuracy
from pyspark.ml.evaluation import MulticlassClassificationEvaluator
evaluator = MulticlassClassificationEvaluator(labelCol="label", predictionCol="prediction", metricName="accuracy")
sparkml_accuracy = evaluator.evaluate(predictions)

# Print off Spark result
print('Spark ml', evaluator.getMetricName(),'=', sparkml_accuracy,", time: %.2f" % sparkml_time, 's')
